"""This is where all the levels are stored"""

import random
import math
import pygame
import pymunk

import serge.actor
import serge.physical
import serge.blocks.utils
import serge.blocks.visualblocks
import serge.events
import serge.sound

from theme import G
import magnet
import goal

class Level(serge.actor.Actor):
    """Base class for all levels"""
    
    order = None
    name = 'level'
    description = 'A level'
    locked = True
    unlocks_levels = []
    completed = False
    
    hint1 = ''
    hint2 = ''
    
    initial_balls = []
    initial_magnets = []
    initial_goals = []
        
    red_target = 100
    green_target = 100
    blue_target = 100

    def __init__(self):
        """Initialise the level"""
        super(Level, self).__init__('level', self.name)
            
    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        self.world = world
        self.scores = scores
        #
        # Balls
        self.balls = []
        self.rotators = []
        self.physics = serge.physical.PhysicalConditions(
                        mass=1.0, radius=32, velocity=(0,0), force=(0,980),
                        elasticity=0.95, layers=2)
        #                        
        for i, (colour, position) in enumerate(self.initial_balls):
            new_ball = self.createBall(colour, position)
        #
        # Magnets
        for colour, position, strength, constraint, minm, maxm in self.initial_magnets:
            m = magnet.Magnet(('%s-magnet' % colour), colour, position, strength, constraint, minm, maxm, scores)
            self.world.addActor(m)
            m.linkEvent(serge.events.E_LEFT_MOUSE_DOWN, screen.clickedMagnet, m)
        #
        self.scores.setTargets(self.red_target, self.green_target, self.blue_target)
        #
        for colour, position in self.initial_goals:
            g = goal.Goal(('%s-goal' % colour), colour, position, self.scores)
            self.world.addActor(g)    

    def createBall(self, colour, position):
        """Create a new ball"""
        new = serge.blocks.utils.addSpriteActorToWorld(self.world, colour, ('%s-b-%d' % (colour, len(self.balls))),
                    ('%s-ball' % colour), 'foreground', position, physics=self.physics.copy())
        #
        # Commented out the sound as it doesn't sound good yet
        #new_ball.linkEvent(serge.events.E_COLLISION, self.bounce)
        self.balls.append(new)
        return new
            
    def getHints(self):
        """Return the hint"""
        hints = []
        for text in (self.hint1, self.hint2):
            hints.append('\n'.join([line.strip() for line in text.splitlines()]))
        return hints
    
    def bounce(self, obj, arg):
        """The ball bounced"""
        serge.sound.Register.playSound('bounce')
        
class Level1(Level):
    order = 1
    name = 'Entry Level'
    description = 'A simple level to get your journey started'
    locked = False
    unlocks_levels = [2]
    
    hint1 = '''Move the balls into
               the coloured zones.
               The score bar increases
               when a ball is
               in the right zone.
               
               You must max out each
               score bar to complete
               the level.'''
               
    hint2 = '''The balls are attracted
               to their coloured
               magnet. Drag the 
               magnet to move it
               around. Use the left
               mouse button to 
               drag.'''
               
    # Balls
    initial_balls = [
        ('red', (300, 500)),
        ('green', (550, 500)),
        ('blue', (800, 500)),
    ]
    
    # Magnets
    initial_magnets = [
        ('red', (300, 650), 10e6, 'y', 100, 990),
        ('green', (550, 670), 5e6, 'y', 100, 990),
        ('blue', (800, 690), 5e6, 'y', 100, 990),
    ]

    initial_goals = [
        ('blue', (300, 490)),
        ('red', (550, 490)),
        ('green', (800, 490)),
    ]
    
    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        super(Level1, self).initLevel(screen, world, scores)


class Level2(Level):
    order = 2
    name = 'Nudging me'
    description = 'Nudge, nudge, wink, wink. Say no more!'
    unlocks_levels = [3]

    hint1 = '''You can nudge a ball
               by right clicking the
               mouse close to it. Hold
               down the right button
               to keep nudging.
               
               The closer you are
               to the ball the more
               the ball will respond.
               You can even
               make it fly!'''
               
    hint2 = '''
               Nudging drains your
               current score. When
               your score is zero 
               you can no longer
               nudge.
               
               Press "HOME"  
               to reset the level.'''
 
    # Balls
    initial_balls = [
        ('red', (480, 300)),
        ('green', (550, 300)),
        ('blue', (620, 300)),
    ]
    
    # Goals
    initial_goals = [
        ('blue', (300, 490)),
        ('red', (550, 310)),
        ('green', (800, 490)),
    ]

    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        #
        # Basically it is level 1
        self.initial_magnets = Level1.initial_magnets
        #
        super(Level2, self).initLevel(screen, world, scores)
        #
        # And a nice horizontal line
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'floor', 'ledge',
                    'long-floor', 'foreground', (550, 400), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=500, height=28)
        )
        #
        # And a tricky little steps
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'floor', 'divider',
                    'short-floor', 'foreground', (550, 552), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=64, height=36)
        )
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'floor', 'left-end',
                    'short-floor', 'foreground', (332, 368), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=64, height=36)
        )
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'floor', 'right-end',
                    'short-floor', 'foreground', (768, 368), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=64, height=36)
        )

class Level3(Level):
    order = 3
    name = 'Never meant'
    description = '... to cause you any sorrow. I only want to see you in the purple drain.'
    unlocks_levels = [4]

    hint1 = '''When the purple ball
               is inside any scoring
               zone it gradually 
               drains the score for
               that zone.
               
               You have to try to
               trap the ball in a
               place where it wont
               cause any harm.'''
               
    hint2 = '''The purple ball is
               not affected by the
               magnets and your
               cannot nudge it!
               
               That doesn't mean you
               cannot move it ... 
               just be creative!'''
 

    # Balls
    initial_balls = [
        ('red', (480, 500)),
        ('green', (550, 500)),
        ('blue', (620, 500)),
        ('purple', (550, 300)),
    ]
    
    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        #
        # Basically it is level 1
        self.initial_magnets = Level1.initial_magnets
        self.initial_goals = Level1.initial_goals
        #
        super(Level3, self).initLevel(screen, world, scores)

class Level4(Level2):
    order = 4
    name = 'All Together'
    description = 'Can I bring my friend to tea?'
    unlocks_levels = [5, 6]

    hint1 = '''As the number of balls
               increases it becomes
               more important to 
               carefully plan out your
               approach.
               
               Especially those pesky
               purple balls!'''
               
    hint2 = '''Press 'HOME' To reset
               the level and try again.
               
               Press 'ESC' to return
               to the main level menu.'''

    # Balls
    initial_balls = [
        ('red', (480, 300)),
        ('green', (550, 300)),
        ('blue', (620, 300)),
        ('purple', (380, 300)),
        ('purple', (720, 300)),
    ]
    


class Level5(Level):
    order = 5
    name = 'Up in the air'
    description = 'Sometimes you need some flair to get some air'
    unlocks_levels = [7, 8]

    hint1 = '''Can you lift it, can
               you hover? 
               
               Can you do it without
               any bother?'''
               
    hint2 = ''''''

    # Balls
    initial_balls = [
        ('red', (480, 450)),
        ('green', (550, 450)),
        ('blue', (620, 450)),
    ]
    
    initial_goals = [
        ('blue', (200, 400)),
        ('red', (550, 490)),
        ('green', (900, 400)),
    ]

    # Magnets
    initial_magnets = [
        ('red', (300, 650), 10e6, 'y', 100, 990),
        ('green', (550, 670), 5e6, 'y', 100, 990),
        ('blue', (800, 690), 5e6, 'y', 100, 990),
        ('blue', (330, 200), 20e6, 'x', 140, 440),
        ('green', (770, 200), 20e6, 'x', 140, 440),
    ]

    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        super(Level5, self).initLevel(screen, world, scores)
        #
        # And a tricky little steps
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'floor', 'left-end',
                    'medium-floor', 'foreground', (200, 500), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=200, height=36)
        )
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'floor', 'right-end',
                    'medium-floor', 'foreground', (900, 500), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=200, height=36)
        )


class Level6(Level):
    order = 6
    name = 'Dig me'
    description = 'It\'s here I sit and rust amid this ruin and rancor like tire irons'
    locked = True
    unlocks_levels = [7, 8]
    
    hint1 = '''I'm ready to leave
               I wanna get out of here
               I'm ready to ride away
               I don't want to die in here.
               
               Dig me ...
               But don't ...
               Bury me.'''
               
    hint2 = ''''''
               
    # Balls
    initial_balls = [
        ('red', (300, 500)),
        ('blue', (800, 500)),
    ]
    
    # Magnets
    initial_magnets = [
        ('red', (300, 650), 10e6, 'y', 100, 990),
        ('green', (550, 670), 5e6, 'y', 100, 990),
        ('blue', (800, 690), 5e6, 'y', 100, 990),
    ]

    initial_goals = [
        ('blue', (300, 490)),
        ('red', (550, 490)),
        ('green', (800, 490)),
    ]
    
    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        super(Level6, self).initLevel(screen, world, scores)
        for i in range(20):
            b = self.createBall('green', (random.randrange(200, 900), random.randrange(200, 400)))
        for i in range(5):
            b = self.createBall('purple', (random.randrange(200, 900), random.randrange(200, 400)))


class Level7(Level):
    order = 7
    name = 'Churn baby'
    description = 'Tumbling like the leaves, yeah we are spiraling on the breeze'
    locked = True
    unlocks_levels = [8, 9, 10]
    
    hint1 = '''Sometimes you just
               have to manage
               the chaos.
               
               Don't try to
               control the
               uncontrollable!'''
               
    hint2 = ''''''
               
    # Balls
    initial_balls = [
        ('red', (300, 500)),
        ('green', (550, 500)),
        ('blue', (800, 500)),
    ]
    
    # Magnets
    initial_magnets = [
        ('red', (300, 650), 10e6, 'y', 100, 990),
        ('green', (550, 670), 5e6, 'y', 100, 990),
        ('blue', (800, 690), 5e6, 'y', 100, 990),
        ('red', (300, 250), 30e6, 'y', 130, 970),
        ('green', (550, 270), 30e6, 'y', 130, 970),
        ('blue', (800, 290), 30e6, 'y', 130, 970),
    ]

    initial_goals = [
        ('blue', (300, 490)),
        ('red', (550, 490)),
        ('green', (800, 490)),
    ]
    
    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        super(Level7, self).initLevel(screen, world, scores)
        left = serge.blocks.utils.addSpriteActorToWorld(self.world, 'ball', 'left-end',
                    'pivoted-medium-floor', 'moving-platforms', (300, 450), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=200, height=36)
        )
        left.getPhysical().body.angular_velocity = 2.0
        self.rotators.append(left)
        right = serge.blocks.utils.addSpriteActorToWorld(self.world, 'ball', 'right-end',
                    'pivoted-medium-floor', 'moving-platforms', (800, 450), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=200, height=36)
        )
        right.getPhysical().body.angular_velocity = -2.0
        self.rotators.append(right)

class Level8(Level):
    order = 8
    name = 'Tip the scales'
    description = 'Promises we plan to break are made in whispered voices'
    locked = True
    unlocks_levels = [9, 10, 11]
    
    hint1 = ''''''
               
    hint2 = ''''''
               
    # Balls
    initial_balls = [
        ('red', (350, 200)),
        ('green', (550, 200)),
        ('blue', (750, 200)),
    ]
    
    # Magnets
    initial_magnets = [
        ('green', (550, 670), 5e6, 'y', 100, 990),
        ('blue', (800, 690), 5e6, 'y', 100, 990),
        ('blue', (800, 200), 5e6, 'y', 130, 970),
    ]

    initial_goals = [
        ('blue', (300, 490)),
        ('red', (550, 350)),
        ('green', (800, 490)),
    ]
    
    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        super(Level8, self).initLevel(screen, world, scores)
        #
        # The main scale
        scale = serge.blocks.utils.addSpriteActorToWorld(self.world, 'scale', 'scale',
                    'pivoted-medium-floor', 'moving-platforms', (550, 350), 
                    physics=serge.physical.PhysicalConditions(
                        group=G('non-collision-group'), mass=10, width=200, height=36, friction=0.1)
        )
        #
        # Pivot it in place
        pivot = pymunk.Body(pymunk.inf, pymunk.inf)
        pivot.position = scale.getPhysical().body.position
        joint = pymunk.PinJoint(scale.getPhysical().body, pivot, (0,0), (0,0))
        scale.getPhysical().space.add(joint)
        #
        # Now bolt magnets on to it
        colour, strength, constraint, minm, maxm, scores = 'red', 30e6, '', 0, 0, self.scores
        self.m1 = m1 = magnet.Magnet(('pivoted-%s-magnet' % colour), colour, 
            (450, 350), strength, constraint, minm, maxm, scores, pivoted=True)
        self.world.addActor(m1)
        bolt = pymunk.PivotJoint(m1.getPhysical().body, scale.getPhysical().body, (0,0), (-100,0))
        m1.getPhysical().space.add(bolt)
        #
        colour, strength, constraint, minm, maxm, scores = 'blue', 10e6, '', 0, 0, self.scores
        self.m2 = m2 = magnet.Magnet(('pivoted-%s-magnet' % colour), colour, 
            (650, 350), strength, constraint, minm, maxm, scores, pivoted=True)
        self.world.addActor(m2)
        bolt = pymunk.PivotJoint(m2.getPhysical().body, scale.getPhysical().body, (0,0), (100,0))
        m2.getPhysical().space.add(bolt)
        #
        self.rotators.extend([scale, m1, m2])

class Level9(Level):
    order = 9
    name = 'The shepherd'
    description = 'Separating the strong from the weak'
    locked = True
    unlocks_levels = [10, 11, 12]
    
    hint1 = '''By moving the
               magnets to and fro
               you can generate
               some momentum.
               
               You can use this to
               spin the platforms
               around.'''
               
    hint2 = '''You need to 
               persevere to
               get all the balls
               in the right
               places!'''
               
    # Balls
    initial_balls = [
        ('red', (350, 200)),
        ('green', (550, 200)),
        ('blue', (750, 200)),
    ]
    
    # Magnets
    initial_magnets = [
        ('red', (530, 650), 20e6, 'y', 100, 990),
        ('green', (550, 670), 5e6, 'y', 100, 990),
        ('blue', (800, 690), 5e6, 'y', 100, 990),
        ('blue', (800, 200), 5e6, 'y', 130, 970),
    ]

    initial_goals = [
        ('blue', (300, 490)),
        ('red', (550, 330)),
        ('green', (800, 490)),
    ]
    
    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        super(Level9, self).initLevel(screen, world, scores)
        #
        # Some pesky purples
        for i in range(0):
            b = self.createBall('purple', (random.randrange(200, 900), random.randrange(200, 400)))
        #
        # And a nice horizontal line
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'floor', 'ledge',
                    'long-floor', 'foreground', (550, 420), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=500, height=28))
        #
        # Some scales
        for name, (x,y) in (('left-scale', (350, 420)), ('right-scale', (750, 420))):
            #
            # The main scale
            scale = serge.blocks.utils.addSpriteActorToWorld(self.world, 'scale', name,
                        'pivoted-medium-floor', 'moving-platforms', (x,y), 
                        physics=serge.physical.PhysicalConditions(
                            group=G('non-collision-group'), mass=10, width=200, height=36, friction=0.1)
            )
            #
            # Pivot it in place
            pivot = pymunk.Body(pymunk.inf, pymunk.inf)
            pivot.position = scale.getPhysical().body.position
            joint = pymunk.PinJoint(scale.getPhysical().body, pivot, (0,0), (0,0))
            scale.getPhysical().space.add(joint)
            #
            # Now bolt magnets on to it
            colour, strength, constraint, minm, maxm, scores = 'red', 30e6, '', 0, 0, self.scores
            self.m1 = m1 = magnet.Magnet(('pivoted-%s-magnet' % colour), colour, 
                (x-100, y), strength, constraint, minm, maxm, scores, pivoted=True)
            self.world.addActor(m1)
            bolt = pymunk.PivotJoint(m1.getPhysical().body, scale.getPhysical().body, (0,0), (-100,0))
            m1.getPhysical().space.add(bolt)
            #
            colour, strength, constraint, minm, maxm, scores = 'blue', 10e6, '', 0, 0, self.scores
            self.m2 = m2 = magnet.Magnet(('pivoted-%s-magnet' % colour), colour, 
                (x+100, y), strength, constraint, minm, maxm, scores, pivoted=True)
            self.world.addActor(m2)
            bolt = pymunk.PivotJoint(m2.getPhysical().body, scale.getPhysical().body, (0,0), (100,0))
            m2.getPhysical().space.add(bolt)
            #
            self.rotators.extend([scale, m1, m2])

class Level10(Level):
    order = 10
    name = 'Poetry in'
    description = '... motion. Creating worlds within worlds'
    locked = True
    unlocks_levels = [10, 11, 12]
    
    hint1 = ''''''
               
    hint2 = ''''''
               
    # Balls
    initial_balls = [
        ('red', (350, 200)),
        ('green', (550, 200)),
        ('blue', (750, 200)),
    ]
    
    # Magnets
    initial_magnets = [
        ('red', (530, 650), 20e6, 'y', 100, 990),
        ('green', (550, 670), 5e6, 'y', 100, 990),
        ('blue', (800, 690), 5e6, 'y', 100, 990),
        ('green', (450, 150), 10e6, 'x', 150, 450),
        ('blue', (650, 150), 10e6, 'x', 150, 450),
    ]

    initial_goals = [
        ('blue', (300, 330)),
        ('red', (550, 330)),
        ('green', (800, 330)),
    ]
    
    def initLevel(self, screen, world, scores):
        """Initialise the level"""
        super(Level10, self).initLevel(screen, world, scores)
        #
        # Some scales
        self.platforms = []
        for name, (x,y) in [('left-platform', (250, 220)), ('middle-platform', (550, 420)), ('middle-platform', (850, 220))]:
            #
            # The main scale
            self.platforms.append(serge.blocks.utils.addSpriteActorToWorld(self.world, 'platform', name,
                        'medium-floor', 'moving-platforms', (x,y), 
                        physics=serge.physical.PhysicalConditions(
                            group=G('non-collision-group'), fixed=True, width=200, height=36, friction=0.1, velocity=(0, 30))
            ))
        #
        spinner = serge.blocks.utils.addSpriteActorToWorld(self.world, 'spinner', 'spinner',
                    'pivoted-medium-floor', 'moving-platforms', (550, 580), 
                    physics=serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=200, height=36)
        )
        spinner.getPhysical().body.angular_velocity = 4.0
        self.rotators.append(spinner)


    def updateActor(self, interval, actor):
        """Update the level"""
        for platform in self.platforms:
            if platform.y > 450 and platform.getPhysical().body.velocity.y > 0:
                platform.getPhysical().body.velocity *= -1
            elif platform.y < 200 and platform.getPhysical().body.velocity.y < 0:
                platform.getPhysical().body.velocity *= -1

def getLevels():
    """Return all the levels"""
    levels = []
    for n, obj in globals().iteritems():
        try:
            if issubclass(obj, Level) and obj is not Level:
                level = obj()
                levels.append((level.order, level))
        except TypeError:
            pass # Not a class
    levels.sort()
    return [l for _, l in levels]
