"""Implementation of a magnet

A magnet is placed around the board and will attract or repel
objects of the same colour. The magnet can have a strength,
which indicates the amount to which is will attract or
repel.

A magnet is also constrained to move on a rail. This means
that it can only move in one direction and can be limited
to the extent that it can move (ie clamped).

"""

import pymunk
import random

import serge.actor
import serge.physical
import serge.visual

from theme import G

class Magnet(serge.actor.CompositeActor):
    """A magnet that attracts or repels objects"""
    
    def __init__(self, name, colour, position, strength, constraint, minm, maxm, scores, pivoted=False):
        """Initialise the Magnet"""
        super(Magnet, self).__init__(colour, name)
        self.colour = colour
        #
        # This magnet can be fixed (user controlled) or pivoting around a point
        if pivoted:
            self.setSpriteName('pivoted-%s-magnet' % self.colour)
            mass = 10
            fixed = False
        else:
            self.setSpriteName('%s-magnet' % self.colour)
            mass = 0
            fixed = True
        #
        self.setLayerName('magnets')
        self.moveTo(*position)
        self.constraint = constraint
        self.minm = minm
        self.maxm = maxm
        self.scores = scores
        self.is_moving = False
        self.strength = strength
        self.mouse = serge.engine.CurrentEngine().getMouse()
        #
        # Physical conditions for the magnet for when it is inside the field of play.
        # Note that we set the group to 2 so that the magnets can go over each other without a problem
        self.setPhysical(serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=fixed, mass=mass,
                         width=70, height=52))
        #
        if self.constraint:
            self.rails = serge.actor.Actor('rails', 'rails')
            self.rails.visual = Rails(self)
            self.rails.setLayerName('background')
            self.addChild(self.rails)
                
    def updateActor(self, interval, world):
        """Update the magnet"""
        #
        # Clear forces on me
        self.getPhysical().body.reset_forces()
        #
        # Attact all the right objects by colour
        for actor in world.findActorsByTag(self.colour):
            if actor is not self:
                v = pymunk.Vec2d(self.x, self.y) - pymunk.Vec2d(actor.x, actor.y)
                f = self.strength/max(10, v.length)**2 * v.normalized()
                actor.getPhysical().body.apply_force(f)
                self.getPhysical().body.apply_force(-f)
        #
        # If we are the one moving then update our position
        if self.is_moving:
            if self.constraint == 'y':
                x = self.mouse.getScreenPos()[0]
                self.x = max(min(x, self.maxm), self.minm)
            elif self.constraint == 'x':
                y = self.mouse.getScreenPos()[1]
                self.y = max(min(y, self.maxm), self.minm)



class Rails(serge.visual.Drawing):
    """Visual representation of the rails that the magnets go along"""

    def __init__(self, parent):
        """Add rails on the screen to show where we are constrained"""
        super(Rails, self).__init__()
        self.parent = parent
        if parent.constraint == 'y':
            self.outer = serge.blocks.visualblocks.Rectangle((parent.maxm-parent.minm, G('rail-outer-size')), G('rail-outer-colour'))
            self.inner = serge.blocks.visualblocks.Rectangle((parent.maxm-parent.minm, G('rail-inner-size')), G('rail-inner-colour'))
        elif parent.constraint == 'x':
            self.outer = serge.blocks.visualblocks.Rectangle((G('rail-outer-size'), parent.maxm-parent.minm), G('rail-outer-colour'))
            self.inner = serge.blocks.visualblocks.Rectangle((G('rail-inner-size'), parent.maxm-parent.minm), G('rail-inner-colour'))
        self.end = serge.blocks.visualblocks.Rectangle((G('rail-end-size'), G('rail-end-size')), G('rail-end-colour'))

            
    def renderTo(self, interval, surface, (x, y)):
        """Render the actor"""
        if self.parent.constraint == 'y':
            self.outer.renderTo(interval, surface, (self.parent.minm, self.parent.y-G('rail-outer-size')/2))
            self.inner.renderTo(interval, surface, (self.parent.minm, self.parent.y-G('rail-inner-size')/2))
            self.end.renderTo(interval, surface, (self.parent.minm, self.parent.y-G('rail-end-size')/2))
            self.end.renderTo(interval, surface, (self.parent.maxm, self.parent.y-G('rail-end-size')/2))
        elif self.parent.constraint == 'x':
            self.outer.renderTo(interval, surface, (self.parent.x-G('rail-outer-size')/2, self.parent.minm))
            self.inner.renderTo(interval, surface, (self.parent.x-G('rail-inner-size')/2, self.parent.minm))
            self.end.renderTo(interval, surface, (self.parent.x-G('rail-end-size')/2, self.parent.minm))
            self.end.renderTo(interval, surface, (self.parent.x-G('rail-end-size')/2, self.parent.maxm))

