"""Main module for the game"""

import pygame
import os
import time

import serge.engine
import serge.world
import serge.actor
import serge.zone
import serge.render
import serge.sound 
try:
    import serge.builder.builder
except ImportError:
    pass

import serge.blocks.utils

import startscreen
import mainscreen
import creditscreen
import levels
import common

from theme import G

def registerSounds():
    """Register the sounds to use"""
    serge.sound.Register.setPath('sound')
    r = serge.sound.Register.registerItem
    r('click', 'letter.wav')
    r('bounce', '51461__andre.rocha__basket_ball_02_bounce.wav')
    
def registerGraphics():
    """Register the graphics to use"""
    serge.visual.Register.setPath('graphics')
    r = serge.visual.Register.registerItem
    rf = serge.visual.Register.registerFromFiles
    r('logo', 'logo.png')
    r('start-screen', 'start-screen.png')
    r('main-screen', 'main-screen.png')
    r('credit-screen', 'credits.png')
    r('light-button', 'light-button.png')
    r('icon', 'icon.png')
    r('red-ball', 'redball.png')
    r('blue-ball', 'blueball.png')
    r('green-ball', 'greenball.png')
    r('purple-ball', 'purpleball.png')
    r('blue-magnet', 'bluemagnet.png')
    r('green-magnet', 'greenmagnet.png')
    r('red-magnet', 'redmagnet.png')
    r('pivoted-blue-magnet', 'pivoted-bluemagnet.png')
    r('pivoted-green-magnet', 'pivoted-greenmagnet.png')
    r('pivoted-red-magnet', 'pivoted-redmagnet.png')
    r('red-goal', 'redgoal.png')
    r('blue-goal', 'bluegoal.png')
    r('green-goal', 'greengoal.png')
    r('bat', 'bat.png')
    r('long-floor', 'long-floor.png')
    r('medium-floor', 'medium-floor.png')
    r('pivoted-medium-floor', 'pivoted-medium-floor.png')
    r('short-floor', 'short-floor.png')
    r('horizontal-wall', 'horizontal-wall.png')
    r('vertical-wall', 'vertical-wall.png')
    rf('level-button', 'level-button-%d.png', 3)
    rf('sound-button', 'sound-%d.png', 2, zoom=0.5)
    rf('level-image', 'snapshot-%d.png', 10, start=0, zoom=0.25)
    r('congratulations-back', 'congratulations-back.png')
    
def startEngine(options):
    """Start the main engine"""
    engine = serge.engine.Engine(width=G('screen-width'), height=G('screen-height'), title=G('screen-title'), icon=G('screen-icon'))
    serge.blocks.utils.createVirtualLayersForEngine(engine, ['background', 'foreground', 'moving-platforms', 'magnets', 'ui'])
    serge.blocks.utils.createWorldsForEngine(engine, ['start-screen', 'main-screen', 'credit-screen'])
    #
    # Link snapshot code in for each world
    if options.snapshot:
        global snapshots
        snapshots = 0
        for world in engine.getWorlds():
            world.linkEvent(serge.events.E_AFTER_RENDER, checkSnapshots, engine.getKeyboard())
    #
    engine.setCurrentWorldByName('start-screen')
    return engine

def toggleSound(obj, (onoff, title)):
    """Toggle whether we are playing sound"""
    common.MUSIC_ON = not common.MUSIC_ON
    onoff.visual.setCell(common.MUSIC_ON)
    if common.MUSIC_ON == 1:
        pygame.mixer.music.pause()
        title.active = False
    else:
        pygame.mixer.music.unpause()
        title.active = True
        
def placeSoundControls(engine, options):
    """Place sound controls on all worlds"""
    #
    # Put volume button on-off button
    onoff = serge.blocks.utils.addSpriteActorToWorld(engine.getWorld('start-screen'), 'persistent', 'onoff', 'sound-button', 'ui',
            center_position=(G('music-onoff-x'), G('music-onoff-y')))
    onoff.visual.setCell(common.MUSIC_ON)
    #
    # And text for the track
    music_title = serge.blocks.utils.addVisualActorToWorld(engine.getWorld('start-screen'), 'persistent', 'music-title', 
            serge.visual.Text('Track', G('music-font-colour'), font_size=G('music-font-size'), justify='left'),
            'ui',
            center_position=(G('music-title-x'), G('music-title-y')))
    #
    credits = serge.blocks.utils.addVisualActorToWorld(engine.getWorld('start-screen'), 'persistent', 'credits', 
        serge.visual.Text('Credits', G('credits-font-colour'), font_size=G('credits-font-size')),
        'ui', center_position=(G('credits-x'), G('credits-y')))
    credits.linkEvent(serge.events.E_LEFT_CLICK, showCredits, (credits, engine))
    #
    onoff.linkEvent(serge.events.E_LEFT_CLICK, toggleSound, (onoff, music_title))
    #music_title.linkEvent(serge.events.E_LEFT_CLICK, toggleSound, (onoff, music_title))
    #
    for world in engine.getWorlds():
        if world.name != 'start-screen':
            world.addActor(onoff)
            world.addActor(music_title)
            world.addActor(credits)
    if options.musicoff:
        toggleSound(None, (onoff, music_title))

def showCredits(obj, (credits, engine)):
    """Show the credits screen"""
    if engine.getCurrentWorld().name != 'credit-screen':
        credits.setText('Back')
        showCredits.back = engine.getCurrentWorld().name
        engine.setCurrentWorldByName('credit-screen')
    else:
        credits.setText('Credits')
        engine.setCurrentWorldByName(showCredits.back)
        
def registerFonts():
    """Set the default fonts"""
    serge.visual.DEFAULT_FONT = os.path.join('graphics', 'Klaudia and Berenika', 'KlaudiaBold', 'KlaudiaBold.ttf')

def checkSnapshots(world, keyboard):
    """Check for doing snapshots"""
    global snapshots
    if keyboard.isClicked(pygame.K_s):
        serge.common.getLogger('Snapshot').info('Taking snapshot %d' % snapshots)
        surface = serge.engine.CurrentEngine().getRenderer().getSurface()
        rect = G('snapshot-%s' % world.name)
        pygame.image.save(surface.subsurface(rect), 'snapshot-%d.png' % snapshots)
        snapshots += 1


def main(options, args):
    """Start the engine and the game"""
    if options.reset:
        common.resetTable()
    common.registerHighScores()
    registerFonts()
    registerSounds()
    registerGraphics()
    #
    # Create the engine
    engine = startEngine(options)
    #
    # Initialise the main logic
    startscreen.main(options)
    mainscreen.main(options)
    creditscreen.main(options)
    #
    placeSoundControls(engine, options)
    common.startMusic('music-start')
    #
    if options.level is not None:
        startscreen.screen.startLevel(None, levels.getLevels()[options.level])
    #
    # Debugging stop at criteria
    if options.end is None:
        end = None
    else:
        end = time.time() + options.end
    #   
    if options.debug:
        serge.builder.builder.main(engine, options.framerate)
    else:
        engine.run(options.framerate, end)
