"""The main screen for the game"""

import random
import math
import pygame
import pymunk

import serge.actor
import serge.visual
import serge.physical
import serge.blocks.utils
import serge.blocks.visualblocks
import serge.events

from theme import G
import magnet
import goal
import scoreboard
import common 

class MainScreen(serge.actor.Actor):
    """The logic for the main screen"""
    
    allow_cheats = False
    
    def __init__(self):
        """Initialise the screen"""
        super(MainScreen, self).__init__('item', 'main-screen')
        self.engine = serge.engine.CurrentEngine()
        self.world = self.engine.getWorld('main-screen')
        self.keyboard = self.engine.getKeyboard()
        self.mouse = self.engine.getMouse()
        self.world.setPhysicsStepsize(1.0)
        
    def initBaseActors(self):
        """Add the basic actors to the level"""
        #
        # Button
        self.next = next = serge.blocks.utils.addVisualActorToWorld(self.world, 'button', 'button', 
            serge.blocks.visualblocks.SpriteText('Play Next Level', G('button-text-colour'), 
                        'light-button', font_size=G('large-text-size')),
            'ui', (550, 150))
        next.linkEvent(serge.events.E_LEFT_CLICK, self.nextLevel)
        next.active = False
        #
        # High score alert
        self.congratulations = serge.blocks.utils.addVisualActorToWorld(self.world, 'alert', 'congratulations', 
            serge.blocks.visualblocks.SpriteText('A new best time!', G('congratulations-text-colour'), 
                'congratulations-back', font_size=G('congratulations-text-size')),
            'ui', (G('congratulations-x'), G('congratulations-y')))
        self.congratulations.active = False
        #
        # bg
        #serge.blocks.utils.addSpriteActorToWorld(self.world, 'bg', 'bg', 'main-screen', 'background')
        #
        # Walls
        self.side_physics = serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=20, height=500)
        self.top_physics = serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=900, height=20)
        self.middle_physics = serge.physical.PhysicalConditions(group=G('non-collision-group'), fixed=True, width=100, height=23)
        #
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'wall', 'left-wall',
                    'vertical-wall', 'foreground', (90, 340), physics=self.side_physics.copy())
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'wall', 'right-wall',
                    'vertical-wall', 'foreground', (1010, 340), physics=self.side_physics.copy())
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'wall', 'top-wall',
                    'horizontal-wall', 'foreground', (550, 100), physics=self.top_physics.copy())
        serge.blocks.utils.addSpriteActorToWorld(self.world, 'wall', 'bottom-wall',
                    'horizontal-wall', 'foreground', (550, 580), physics=self.top_physics.copy())
        #
        # Set up the goals
        self.scores = scoreboard.ScoreBoard(1,1,1)
        self.world.addActor(self.scores)
        #
        # The labels for the level
        self.title = serge.blocks.utils.addVisualActorToWorld(self.world, 'text', 'title', 
            serge.visual.Text('Level', G('title-text-colour'), font_size=G('title-text-size')),
            'ui', (550, G('title-text-y')))
        self.subtitle = serge.blocks.utils.addVisualActorToWorld(self.world, 'text', 'subtitle', 
            serge.visual.Text('The sub title for the level', G('subtitle-text-colour'), font_size=G('subtitle-text-size')),
            'ui', (550, G('subtitle-text-y')))
        self.hint1 = serge.blocks.utils.addVisualActorToWorld(self.world, 'text', 'hint1', 
            serge.visual.Text('', G('hint1-text-colour'), font_size=G('hint1-text-size'), justify='left'),
            'ui', (G('hint1-text-x'), G('hint1-text-y')))
        self.hint2 = serge.blocks.utils.addVisualActorToWorld(self.world, 'text', 'hint2', 
            serge.visual.Text('', G('hint2-text-colour'), font_size=G('hint2-text-size'), justify='left'),
            'ui', (G('hint2-text-x'), G('hint2-text-y')))
        #
        self.moving = None
        self.complete = False
        self.just_clicked = False
        #
        self.world.addActor(self)

    def initLevel(self, level):
        """Initialize our state with the given level"""
        #
        # Clear the world 
        self.world.clearActorsExceptTags(['persistent'])
        self.initBaseActors()
        level.initLevel(self, self.world, self.scores)
        best = common.HIGH_SCORES.getCategory('level %d' % level.order)
        self.scores.best = best[0][1] if best else None
        self.world.addActor(level)
        self.level = level
        self.title.setText(level.name)
        self.subtitle.setText(level.description)
        self.hint1.setText(level.getHints()[0])
        self.hint2.setText(level.getHints()[1])
        
    def nextLevel(self, obj, arg):
        """Move on to the next level"""
        serge.sound.Register.playSound('click')
        self.log.info('Clicked next level')
        world = serge.engine.CurrentEngine().setCurrentWorldByName('start-screen')
        common.startMusic('music-start')
               
        
    def updateActor(self, world, interval):
        """Update this actor"""
        for ball in self.level.balls + self.level.rotators:
            ball.setAngle(math.degrees(-ball.getPhysical().body.angle))
        #
        # Look for escape
        if self.keyboard.isClicked(pygame.K_ESCAPE):
            serge.sound.Register.playSound('click')
            world = self.engine.setCurrentWorldByName('start-screen')
            common.startMusic('music-start')
        if self.keyboard.isClicked(pygame.K_HOME):
            serge.sound.Register.playSound('click')
            self.initLevel(self.level)
            return            
        #
        # Look for end of game
        if not self.complete and self.scores.targetComplete():
            self.log.info('The level is complete')
            self.complete = True
            self.next.active = True
            self.level.completed = True
            result = common.HIGH_SCORES.addScore('level %d' % self.level.order, 'player', self.scores.clock)
            if result:
                self.congratulations.active = True
            common.saveScores()
            self.log.info('Result of adding score is %s' % result)
        #
        # Cheat mode
        if self.allow_cheats and self.keyboard.isClicked(pygame.K_c):
            self.scores.setComplete()
        #
        # Unselect magnet if we need to
        if self.mouse.isUp(serge.input.M_LEFT) and self.moving:
            self.log.debug('Just clicked')
            self.moving.is_moving = False
            self.moving = None
            self.just_clicked = True

        
    def clickedMagnet(self, obj, arg):
        """Clicked on a magnet"""
        if not self.moving and not arg.is_moving:
            self.log.debug('Clicked on %s' % (arg.getNiceName()))
            arg.is_moving = True
            self.moving = arg


    def renderTo(self, interval, renderer):
        """Render this actor"""
        #
        # We are abusing this method to reset the forces on all the balls
        # The renderTo method gets called after all the physics has happened
        # so we know that it is safe to do stuff here
        g = pymunk.Vec2d(0, 980)
        #
        # Set forces on the ball
        for ball in self.level.balls:
            #
            # Randomizer force which operates when right clicking
            if ball.tag != 'purple' and self.mouse.isDown(serge.input.M_RIGHT) and not self.scores.isZero():
                v = self.mouse.getScreenPos() - pymunk.Vec2d(ball.x, ball.y)
                m = -1e6/(max(10, v.length)**2) * v.normalized()
            else:
                m = pymunk.Vec2d(0,0)
            #
            ball.getPhysical().body.reset_forces()
            ball.getPhysical().body.apply_force(g+m)
            
def main(options):
    """Create the main logic"""
    s = MainScreen()
    s.world.addActor(s)
    s.allow_cheats = options.cheat

