"""The initial screen for the game"""

import pygame
import random
import math
import pymunk
import os

import serge.actor
import serge.events
import serge.blocks.utils
import serge.blocks.visualblocks
import serge.blocks.layout

from theme import G
import common 
import levels

class StartScreen(serge.actor.Actor):
    """The logic for the start screen"""
    
    allow_cheats = False
    
    def __init__(self):
        """Initialise the screen"""
        self.engine = serge.engine.CurrentEngine()
        self.world = self.engine.getWorld('main-screen')
        self.keyboard = self.engine.getKeyboard()
        self.mouse = self.engine.getMouse()
        super(StartScreen, self).__init__('item', 'start-screen')
        #
        self.music_pos = 0
        #
        self.world = serge.engine.CurrentEngine().getWorld('start-screen')
        #
        # bg
        #serge.blocks.utils.addSpriteActorToWorld(self.world, 'bg', 'bg', 'start-screen', 'background')
        logo = serge.blocks.utils.addSpriteActorToWorld(self.world, 'bg', 'bg', 'logo', 'ui')
        logo.y = G('logo-y')
        #
        self.version = serge.blocks.utils.addVisualActorToWorld(self.world, 'version', 'version', 
            serge.visual.Text('v'+common.version, G('version-colour'), font_size=G('version-size'), justify='left'),
            'ui', center_position=(G('version-x'), G('version-y')))
        #
        # The grid to show the levels
        self.grid = serge.blocks.layout.Grid('grid', 'grid', height=G('level-grid-height'), size=(G('level-grid-nx'), G('level-grid-ny')))
        self.grid.setLayerName('ui')
        self.grid.y = G('level-grid-y')
        self.levels = []
        for i, level in enumerate(levels.getLevels()):
            text = level.name
            button = serge.actor.Actor('level-button', ('l-%d' % i)) 
            button.visual = serge.blocks.visualblocks.TextToggle(text, G('level-text-colour'), 
                                'level-button', font_size=G('level-title-size'))
            button.linkEvent(serge.events.E_LEFT_CLICK, self.startLevel, level)
            self.grid.addActor((i%G('level-grid-nx'), i//G('level-grid-nx')), button)
            #
            # Back image
            image = serge.blocks.utils.addSpriteActorToWorld(self.world, 'level', 'level', 'level-image', 'foreground')
            image.visual.setCell(i)
            image.x = button.x
            image.y = button.y
            #
            self.levels.append((level, button))
        self.world.addActor(self.grid)
        #
        self.initVisuals()
        
    def startLevel(self, obj, level):
        """Clicked on a level to start"""
        if self.allow_cheats or not level.locked:
            self.log.info('Starting level %s' % level)
            serge.sound.Register.playSound('click')
            #
            # Store music position for restarting later
            self.music_pos = pygame.mixer.music.get_pos()
            #
            # And start the relevant music
            common.startMusic('music-%d' % ((level.order-1)%G('music-max-tunes')))
            #
            self.log.info('Clicked start for level %s' % level)
            world = serge.engine.CurrentEngine().setCurrentWorldByName('main-screen')
            main = world.findActorByName('main-screen').initLevel(level)
        else:
            self.log.info('Level %s is locked' % level)

    def updateActor(self, interval, world):
        """Update the actor"""
        #
        # Look for escape
        if self.keyboard.isClicked(pygame.K_ESCAPE):
            serge.sound.Register.playSound('click')
            self.engine.stop()
        #
        # Check for completed
        for level, _ in self.levels:
            if level.completed:
                for n in level.unlocks_levels:
                    for other, _ in self.levels:
                        if other.order == n:
                            other.locked = False
        #
        # Check for locked
        for level, button in self.levels:
            if level.completed:
                button.visual.setCell(2)
                button.visual.text_visual.setColour(G('level-text-completed-colour'))
            elif not level.locked:
                button.visual.setCell(0)
                button.visual.text_visual.setColour(G('level-text-colour'))
            else:
                button.visual.setCell(1)
                button.visual.text_visual.setColour(G('level-text-locked-colour'))
        #
        # Set ball rotations
        for ball in self.balls:
            ball.setAngle(math.degrees(-ball.getPhysical().body.angle))
        #
        # Check for switch of gravity
        if True or random.random() < G('start-gravity-switch'):
            g = pymunk.Vec2d(0,random.randrange(0, G('start-gravity'))).rotated(random.randrange(0, 360))
            for ball in self.balls:
                v = pymunk.Vec2d(self.mouse.getScreenPos()) - pymunk.Vec2d(ball.x, ball.y)
                f = 20e6/max(10, v.length)**2 * v.normalized()
                ball.getPhysical().body.force = g-f
            
    def initVisuals(self):
        """Initialise the background visual elements"""
        physics = serge.physical.PhysicalConditions(
                        mass=1.0, radius=32, velocity=(0,0), force=(0,G('start-gravity')),
                        elasticity=0.85, layers=2)
        #                        
        self.balls = []
        for i in range(G('start-ball-number')):
            colour = random.choice(('red', 'green', 'blue', 'purple'))
            position = (random.randrange(0, G('screen-width')), random.randrange(0, G('screen-height')))
            new_ball = serge.blocks.utils.addSpriteActorToWorld(self.world, colour, ('%s-b-%d' % (colour, i)),
                    ('%s-ball' % colour), 'background', position, physics=physics.copy())
            self.balls.append(new_ball)
        #
        side_physics = serge.physical.PhysicalConditions(fixed=True, width=20, height=G('screen-height'))
        top_physics = serge.physical.PhysicalConditions(fixed=True, width=G('screen-width'), height=20)
        #
        colour = G('start-colour')
        serge.blocks.utils.addVisualActorToWorld(self.world, 'wall', 'left-wall',
                    serge.blocks.visualblocks.Rectangle((20, G('screen-height')), colour), 
                    'background', (0, G('screen-height')/2), physics=side_physics.copy())
        serge.blocks.utils.addVisualActorToWorld(self.world, 'wall', 'right-wall',
                    serge.blocks.visualblocks.Rectangle((20, G('screen-height')), colour), 
                    'background', (G('screen-width'), G('screen-height')/2), physics=side_physics.copy())
        serge.blocks.utils.addVisualActorToWorld(self.world, 'wall', 'top-wall',
                    serge.blocks.visualblocks.Rectangle((G('screen-width'), 20), colour), 
                    'background', (G('screen-width')/2, 0), physics=top_physics.copy())
        serge.blocks.utils.addVisualActorToWorld(self.world, 'wall', 'bottom-wall',
                    serge.blocks.visualblocks.Rectangle((G('screen-width'), 20), colour), 
                    'background', (G('screen-width')/2, G('screen-height')), physics=top_physics.copy())
        #
        # Experimental blur
        def doRender(surface):
            s = blurSurf(l.surface, 10.0)
            surface.blit(s, (0,0))

        l = self.engine.getRenderer().getLayer('background')
        #l.render = doRender
        
        
        
def blurSurf(surface, amt):
    """
    Blur the given surface by the given 'amount'.  Only values 1 and greater
    are valid.  Value 1 = no blur.
    """
    if amt < 1.0:
        raise ValueError("Arg 'amt' must be greater than 1.0, passed in value is %s"%amt)
    scale = 1.0/float(amt)
    surf_size = surface.get_size()
    scale_size = (int(surf_size[0]*scale), int(surf_size[1]*scale))
    surf = pygame.transform.smoothscale(surface, scale_size)
    surf = pygame.transform.smoothscale(surf, surf_size)
    return surf        


def main(options):
    """Create the main logic"""
    global screen                
    screen = StartScreen()
    screen.world.addActor(screen)
    screen.allow_cheats = options.cheat
